/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "paxos.h"
#include <fstream>
#include <iostream>
#include <assert.h>

paxos_log::paxos_log(paxos *_pxs, std::string _me, std::string log_dir)
  : pxs (_pxs)
{
  name = log_dir + "/paxos-" + _me + ".log";
  logread();
}

std::vector<std::string>
paxos_log::logread_v(std::ifstream &from)
{
  int n;
  std::vector<std::string> v;
  std::string m;

  from >> n;
  for (int i = 0; i < n; i++) {
    from >> m;
    v.push_back(m);
  }
  return v;
}

void
paxos_log::log_v(std::ofstream &f, std::vector<std::string> v)
{
  f << v.size();
  f << " ";
  for (unsigned i = 0; i < v.size(); i++) {
    f << v[i];
    f << " ";
  }
}

void
paxos_log::logread(void)
{
  std::ifstream from;
  std::string type;
  unsigned vid;

  from.open(name.c_str());
  wfs_log(WFS_DBG_4,"logread called\n");
  while (from >> type) {
    if (type == "view") {
      from >> vid;
      pxs->views[vid] = logread_v(from);
      pxs->vid_h = vid;
      wfs_log(WFS_DBG_4,"logread: vid: %d w. v = %s\n", vid,
			pxs->print_members(pxs->views[vid]).c_str());
      pxs->c_vid = vid;
      pxs->c_mems = pxs->views[vid];
      pxs->v_a.clear();
      pxs->n_h.n = 0;
      pxs->n_a.n = 0;
    } else if (type == "high") {
      from >> pxs->n_h.n;
      from >> pxs->n_h.m;
      wfs_log(WFS_DBG_4,"logread: high update: %d(%s)\n", pxs->n_h.n, pxs->n_h.m.c_str());
    } else if (type == "prop") {
      from >> pxs->n_a.n;
      from >> pxs->n_a.m;
      pxs->v_a = logread_v(from);
      wfs_log(WFS_DBG_4,"logread: prop update %d(%s) with v = %s\n", pxs->n_a.n,
	     pxs->n_a.m.c_str(), pxs->print_members(pxs->v_a).c_str());
    } else {
      wfs_log(WFS_DBG_4,"logread: unknown log record\n");
      assert(0);
    }
  }
  from.close();
}

// XXX should be an atomic operation
void
paxos_log::logview(unsigned vid, std::vector<std::string> v)
{
  std::ofstream f;
  f.open(name.c_str(), std::ios::app);
  f << "view";
  f << " ";
  f << vid;
  f << " ";
  log_v(f, v);
  f << "\n";
  f.close();
}

void
paxos_log::loghigh(prop_t n_h)
{
  std::ofstream f;
  f.open(name.c_str(), std::ios::app);
  f << "high";
  f << " ";
  f << n_h.n;
  f << " ";
  f << n_h.m;
  f << "\n";
  f.close();
}

void
paxos_log::logprop(prop_t n, std::vector<std::string> v)
{
  std::ofstream f;
  f.open(name.c_str(), std::ios::app);
  f << "prop";
  f << " ";
  f << n.n;
  f << " ";
  f << n.m;
  f << " ";
  log_v(f, v);
  f << "\n";
  f.close();
}

