/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef rsm_protocol_h
#define rsm_protocol_h
#include <sstream>
#include <string>
#include "rpc.h"

class rsm_client_protocol {
 public:
  enum xxstatus { OK, ERR, NOTMASTER, BUSY};
  typedef int status;
  enum rpc_numbers {
    invoke = 0x9001,
    members,
  };
};

class rsm_protocol {
 public:
  enum xxstatus { OK, ERR, BUSY};
  typedef int status;
  enum rpc_numbers {
    invoke = 0x10001,
    transferreq,
    logtransferreq,
    cptransferreq,
    viewstampreq,
  };
};

struct viewstamp {
  viewstamp (unsigned int _vid = 0, unsigned int _seqno = 0) {
    vid = _vid;
    seqno = _seqno;
  };
  unsigned int vid;
  unsigned int seqno;
};

struct rsm_record {
rsm_record(): vs(0,0), procno(0), cp(false) {};
	rsm_record(const viewstamp &v, int p, const std::string &s) {
		vs = v;
		procno = p;
		req = s;
		cp = false;
	};
	viewstamp vs;
	int procno;
	std::string req;
  uint16_t cp; // (bool)
};


inline bool operator==(viewstamp a, viewstamp b) {
  return a.vid == b.vid && a.seqno == b.seqno;
}

inline bool operator>(viewstamp a, viewstamp b) {
  return (a.vid > b.vid) || ((a.vid == b.vid) && a.seqno > b.seqno);
}

inline bool operator<(viewstamp a, viewstamp b) {
  return (a.vid < b.vid) || ((a.vid == b.vid) && a.seqno < b.seqno);
}

inline bool operator<=(viewstamp a, viewstamp b) {
  return (a < b) || (a == b);
}

inline bool operator!=(viewstamp a, viewstamp b) {
  return a.vid != b.vid || a.seqno != b.seqno;
}

inline marshall& operator<<(marshall &m, const viewstamp &v)
{
  m << v.vid;
  m << v.seqno;
  return m;
}

inline unmarshall& operator>>(unmarshall &u, viewstamp &v) {
  u >> v.vid;
  u >> v.seqno;
  return u;
}

inline marshall& operator<<(marshall &m, const rsm_record &r)
{
	m << r.vs;
	m << r.procno;
	m << r.req;
        m << r.cp;
	return m;
}

inline unmarshall& operator>>(unmarshall &u, rsm_record &r)
{
	u >> r.vs;
	u >> r.procno;
	u >> r.req;
	u >> r.cp;
	return u;
}

#endif
