/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __WFS_CLIENT_MANAGER_H__
#define __WFS_CLIENT_MANAGER_H__ 1

#include "wfs_file.h"
#include "wfs_node.h"
#include <map>
#include <rpc.h>

class wfs_client_manager {

 private:

  struct client_node {
    wfs_node_id id;
    rpcc *client;
    int out;
    client_node *prev;
    client_node *next;
  };

  struct binding {
    int waiters;
    pthread_cond_t *cond;
  };

  vivaldi *_viv;
  std::map< wfs_node_id, client_node *> _clients;
  client_node *_head;
  client_node *_tail;
  pthread_mutex_t _clients_lock;
  std::map< wfs_node_id, binding *> _bindings;

  static const unsigned CLIENT_LIMIT;

  bool delete_client( client_node *cnode );

 public:
  wfs_client_manager( vivaldi *viv );
  ~wfs_client_manager();
  rpcc *get_client( const wfs_node &server, rpcc::TO to = WFS_DEFAULT_TO );
  void done_with_client( const wfs_node &server );
  
};

#endif // __WFS_CLIENT_MANAGER_H__
