/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __WFS_NODE_H__
#define __WFS_NODE_H__ 1

#include <map>
#include <string>
#include "host.h"

// Get macros for printf/scanfing fixed-size integral types
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#include <netinet/in.h>

// Various IDs
typedef uint64_t wfs_node_id;

struct wfs_node {

  wfs_node() : id(0), site_id("") {}
  wfs_node_id id;
  sockaddr_in addr;
  std::string site_id;

};

inline
wfs_node_id two_ints_to_node_id( uint32_t nid_1, uint32_t nid_2 ) {
  wfs_node_id nid = nid_1;
  nid <<= sizeof(uint32_t)*8;
  nid += nid_2;
  return nid;
}

wfs_node make_node( const char *ip, unsigned int port, 
		    std::string site_id = "" );

wfs_node make_node( const char *ipandport,
		    std::string site_id = "" );

wfs_node make_node( host h );

#endif // __WFS_NODE_H__
